from __future__ import absolute_import
from __future__ import print_function

import math

import veriloggen.core.vtypes as vtypes
import veriloggen.types.ram as ram
import veriloggen.types.fifo as fifo
import veriloggen.types.axi as axi
from veriloggen.seq.seq import Seq

from . import compiler

def Lock(m, name, clk, rst, width=32):
    """ alias of Mutex class """
    return Mutex(m, name, clk, rst, width)


class Mutex(object):
    __intrinsics__ = ('lock', 'try_lock', 'unlock',
                      'acquire', 'release')

    def __init__(self, m, name, clk, rst, width=32):

        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.width = width

        self.seq = Seq(self.m, self.name, self.clk, self.rst)

        self.lock_reg = self.m.Reg(
            '_'.join(['', self.name, 'lock_reg']), initval=0)
        self.lock_id = self.m.Reg(
            '_'.join(['', self.name, 'lock_id']), self.width, initval=0)

        self.id_map = {}
        self.id_map_count = 0

    def lock(self, fsm):
        name = fsm.name
        new_lock_id = self._get_id(name)

        if new_lock_id > 2 ** self.width - 1:
            raise ValueError('too many lock IDs')

        # try
        try_state = fsm.current

        state_cond = fsm.state == fsm.current
        try_cond = vtypes.Not(self.lock_reg)
        fsm_cond = vtypes.Ors(try_cond, self.lock_id == new_lock_id)

        self.seq.If(state_cond, try_cond)(
            self.lock_reg(1),
            self.lock_id(new_lock_id)
        )

        fsm.If(fsm_cond).goto_next()

        # verify
        cond = vtypes.Ands(self.lock_reg, self.lock_id == new_lock_id)
        fsm.If(vtypes.Not(cond)).goto(try_state)  # try again
        fsm.If(cond).goto_next()  # OK

        return 1

    def try_lock(self, fsm):
        name = fsm.name
        new_lock_id = self._get_id(name)

        if new_lock_id > 2 ** self.width - 1:
            raise ValueError('too many lock IDs')

        # try
        try_state = fsm.current

        state_cond = fsm.state == fsm.current
        try_cond = vtypes.Not(self.lock_reg)

        self.seq.If(state_cond, try_cond)(
            self.lock_reg(1),
            self.lock_id(new_lock_id)
        )

        fsm.goto_next()

        # verify
        cond = vtypes.And(self.lock_reg, self.lock_id == new_lock_id)
        result = self.m.TmpReg(initval=0)
        fsm(
            result(cond)
        )
        fsm.goto_next()

        return result

    def unlock(self, fsm):
        name = fsm.name
        new_lock_id = self._get_id(name)

        if new_lock_id > 2 ** self.width - 1:
            raise ValueError('too many lock IDs')

        state_cond = fsm.state == fsm.current

        self.seq.If(state_cond, self.lock_id == new_lock_id)(
            self.lock_reg(0)
        )

        fsm.goto_next()

        return 0

    def _get_id(self, name):
        if name not in self.id_map:
            self.id_map[name] = self.id_map_count
            self.id_map_count += 1

        return self.id_map[name]

    def acquire(self, fsm, blocking=True):
        """ alias of lock() """

        if not isinstance(blocking, (bool, int)):
            raise TypeError('blocking argument must be bool')

        if blocking:
            return self.lock(fsm)

        return self.try_lock(fsm)

    def release(self, fsm):
        """ alias of unlock() """

        return self.unlock(fsm)


class Barrier(object):
    __intrinsics__ = ('wait', )

    def __init__(self, m, name, clk, rst, numparties):

        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.numparties = numparties
        self.width = int(math.ceil(math.log(self.numparties, 2))) + 1

        self.seq = Seq(self.m, self.name, self.clk, self.rst)

        self.count = self.m.Reg(
            '_'.join(['', self.name, 'barrier_count']), self.width, initval=0)
        self.done = self.m.Reg(
            '_'.join(['', self.name, 'barrier_done']), initval=0)
        self.mutex = Mutex(self.m, '_'.join(
            ['', self.name, 'barrier_mutex']), self.clk, self.rst)

        # reset condition
        self.seq(
            self.done(0)
        )
        self.seq.If(self.count == self.numparties)(
            self.count(0),
            self.done(1)
        )

    def wait(self, fsm):

        self.mutex.lock(fsm)

        state_cond = fsm.state == fsm.current
        self.seq.If(state_cond)(
            self.count.inc()
        )
        fsm.goto_next()

        self.mutex.unlock(fsm)

        fsm.If(self.done).goto_next()

        return 0


class Shared(object):
    __intrinsics__ = ('read', 'write',
                      'lock', 'try_lock', 'unlock')

    def __init__(self, value):
        self._value = value
        self.seq = None
        self.mutex = None

    @property
    def value(self):
        return self._value

    def read(self, fsm):
        return self._value

    def write(self, fsm, value, *part):
        if self.seq is None:
            self.seq = Seq(fsm.m, '_'.join(
                ['seq', self._value.name]), fsm.clk, fsm.rst)

        cond = fsm.state == fsm.current

        def getval(v, p):
            if isinstance(p, (tuple, list)):
                return v[p[0], p[1]]
            return v[p]

        if len(part) == 0:
            targ = self._value
        elif len(part) == 1:
            targ = getval(self._value, part[0])
        elif len(part) == 2:
            targ = getval(getval(self._value, part[0]), part[1])
        else:
            raise TypeError('unsupported type')

        self.seq.If(cond)(
            targ(value)
        )

        fsm.goto_next()
        return 0

    def lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.lock(fsm)

    def try_lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.try_lock(fsm)

    def unlock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.unlock(fsm)


class RAM(ram.SyncRAMManager):
    __intrinsics__ = ('read', 'write', 'dma_read', 'dma_write',
                      'lock', 'try_lock', 'unlock')

    def __init__(self, m, name, clk, rst,
                 datawidth=32, addrwidth=10, numports=1):

        ram.SyncRAMManager.__init__(
            self, m, name, clk, rst, datawidth, addrwidth, numports)

        self.mutex = None

    def read(self, fsm, addr, port=0, unified=False):
        """ intrinsic read operation """

        if unified:
            return self._unified_read(fsm, addr, port)

        return self._shared_read(fsm, addr, port)

    def write(self, fsm, addr, wdata, port=0, unified=False):
        """ intrinsic write operation """

        if unified:
            return self._unified_write(fsm, addr, wdata, port)

        return self._shared_write(fsm, addr, wdata, port)

    def _shared_read(self, fsm, addr, port=0):
        """ intrinsic read operation using a shared Seq object """

        cond = fsm.state == fsm.current

        rdata, rvalid = ram.SyncRAMManager.read(self, port, addr, cond)
        rdata_reg = self.m.TmpReg(self.datawidth, initval=0, signed=True)

        fsm.If(rvalid)(
            rdata_reg(rdata)
        )
        fsm.Then().goto_next()

        return rdata_reg

    def _shared_write(self, fsm, addr, wdata, port=0):
        """ intrinsic write operation using a shared Seq object """

        cond = fsm.state == fsm.current

        ram.SyncRAMManager.write(self, port, addr, wdata, cond)
        fsm.goto_next()

        return 0

    def _unified_read(self, fsm, addr, port=0):
        """ intrinsic read operation using the given FSM object """

        fsm(
            self.interfaces[port].addr(addr)
        )

        for _ in range(2):
            fsm.goto_next()

        rdata = self.m.TmpReg(self.datawidth, initval=0, signed=True)

        fsm(
            rdata(self.interfaces[port].rdata)
        )
        fsm.goto_next()

        return rdata

    def _unified_write(self, fsm, addr, wdata, port=0):
        """ intrinsic write operation using the given FSM object """

        if self._write_disabled[port]:
            raise TypeError('Write disabled.')

        fsm(
            self.interfaces[port].addr(addr),
            self.interfaces[port].wdata(wdata),
            self.interfaces[port].wenable(1)
        )
        fsm.Delay(1)(
            self.interfaces[port].wenable(0)
        )
        fsm.goto_next()

        return 0

    def dma_read(self, fsm, bus, local_addr, global_addr, size, port=0):
        if not isinstance(bus, AXIM):
            raise TypeError('AXIM interface is required')

        return bus.dma_read(fsm, self, local_addr, global_addr, size, port)

    def dma_write(self, fsm, bus, local_addr, global_addr, size, port=0):
        if not isinstance(bus, AXIM):
            raise TypeError('AXIM interface is required')

        return bus.dma_write(fsm, self, local_addr, global_addr, size, port)

    def lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.lock(fsm)

    def try_lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.try_lock(fsm)

    def unlock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.unlock(fsm)


class FIFO(fifo.Fifo):
    __intrinsics__ = ('enq', 'deq', 'try_enq', 'try_deq',
                      'is_empty', 'is_almost_empty',
                      'is_full', 'is_almost_full',
                      'lock', 'try_lock', 'unlock')

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=4):
        fifo.Fifo.__init__(self, m, name, clk, rst, datawidth, addrwidth)
        self.mutex = None

    def enq(self, fsm, wdata):
        cond = fsm.state == fsm.current

        ack, ready = fifo.Fifo.enq(self, wdata, cond=cond)
        fsm.If(ready).goto_next()

        return 0

    def deq(self, fsm):
        cond = fsm.state == fsm.current

        rdata, rvalid = fifo.Fifo.deq(self, cond=cond)
        fsm.If(vtypes.Not(self.empty)).goto_next()
        fsm.goto_next()

        rdata_reg = self.m.TmpReg(self.datawidth, initval=0, signed=True)

        fsm.If(rvalid)(
            rdata_reg(rdata)
        )
        fsm.If(rvalid).goto_next()

        return rdata_reg

    def try_enq(self, fsm, wdata):
        cond = fsm.state == fsm.current

        ack, ready = fifo.Fifo.enq(self, wdata, cond=cond)
        fsm.goto_next()

        ack_reg = self.m.TmpReg(initval=0)
        fsm(
            ack_reg(ack)
        )
        fsm.goto_next()

        return ack_reg

    def try_deq(self, fsm):
        cond = fsm.state == fsm.current

        rdata, rvalid = fifo.Fifo.deq(self, cond=cond)
        fsm.goto_next()
        fsm.goto_next()

        rdata_reg = self.m.TmpReg(self.datawidth, initval=0, signed=True)
        rvalid_reg = self.m.TmpReg(initval=0)

        fsm(
            rdata_reg(rdata),
            rvalid_reg(rvalid)
        )
        fsm.goto_next()

        return rdata_reg, rvalid_reg

    def is_almost_empty(self, fsm):
        fsm.goto_next()
        return self.almost_empty

    def is_empty(self, fsm):
        fsm.goto_next()
        return self.empty

    def is_almost_full(self, fsm):
        fsm.goto_next()
        return self.almost_full

    def is_full(self, fsm):
        fsm.goto_next()
        return self.full

    def lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.lock(fsm)

    def try_lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.try_lock(fsm)

    def unlock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.unlock(fsm)


class AXIM(axi.AxiMaster):
    __intrinsics__ = ('dma_read', 'dma_write',
                      'lock', 'try_lock', 'unlock')

    burstlen = 256

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=32):
        axi.AxiMaster.__init__(self, m, name, clk, rst, datawidth, addrwidth)
        self.mutex = None

    def dma_read(self, fsm, ram, local_addr, global_addr, size, port=0):
        if not isinstance(ram, RAM):
            raise TypeError('RAM is required')

        req_local_addr = self.m.TmpReg(ram.addrwidth, initval=0)
        req_global_addr = self.m.TmpReg(self.addrwidth, initval=0)
        req_size = self.m.TmpReg(self.addrwidth, initval=0)
        rest_size = self.m.TmpReg(self.addrwidth, initval=0)

        fsm(
            req_local_addr(local_addr),
            req_global_addr(global_addr),
            rest_size(size)
        )
        fsm.goto_next()

        check_state = fsm.current
        fsm.If(rest_size <= self.burstlen)(
            req_size(rest_size),
            rest_size(0)
        ).Else(
            req_size(self.burstlen),
            rest_size(rest_size - self.burstlen)
        )
        fsm.goto_next()

        cond = fsm.state == fsm.current

        done = axi.AxiMaster.dma_read(self, ram, req_global_addr, req_local_addr, req_size,
                                      cond=cond, ram_port=port)

        fsm.If(done)(
            req_local_addr.add(req_size),
            req_global_addr.add(compiler.optimize(req_size * (self.datawidth // 8)))
        )
        fsm.If(done, rest_size > 0).goto(check_state)
        fsm.If(done, rest_size == 0).goto_next()

        return 0

    def dma_write(self, fsm, ram, local_addr, global_addr, size, port=0):
        if not isinstance(ram, RAM):
            raise TypeError('RAM is required')

        req_local_addr = self.m.TmpReg(ram.addrwidth, initval=0)
        req_global_addr = self.m.TmpReg(self.addrwidth, initval=0)
        req_size = self.m.TmpReg(self.addrwidth, initval=0)
        rest_size = self.m.TmpReg(self.addrwidth, initval=0)

        fsm(
            req_local_addr(local_addr),
            req_global_addr(global_addr),
            rest_size(size)
        )
        fsm.goto_next()

        check_state = fsm.current
        fsm.If(rest_size <= self.burstlen)(
            req_size(rest_size),
            rest_size(0)
        ).Else(
            req_size(self.burstlen),
            rest_size(rest_size - self.burstlen)
        )
        fsm.goto_next()

        cond = fsm.state == fsm.current

        done = axi.AxiMaster.dma_write(self, ram, req_global_addr, req_local_addr, req_size,
                                       cond=cond, ram_port=port)

        fsm.If(done)(
            req_local_addr.add(req_size),
            req_global_addr.add(compiler.optimize(req_size * (self.datawidth // 8)))
        )
        fsm.If(done, rest_size > 0).goto(check_state)
        fsm.If(done, rest_size == 0).goto_next()

        return 0

    def lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.lock(fsm)

    def try_lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.try_lock(fsm)

    def unlock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.unlock(fsm)


class AXIS(axi.AxiSlave):
    __intrinsics__ = ('lock', 'try_lock', 'unlock')

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=32):
        axi.AxiSlave.__init__(self, m, name, clk, rst, datawidth, addrwidth)
        self.mutex = None

    def lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.lock(fsm)

    def try_lock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.try_lock(fsm)

    def unlock(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

        return self.mutex.unlock(fsm)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *
import veriloggen.dataflow as dataflow

def mux(cond, x, y):
    if dataflow.is_dataflow_object(cond, x, y):
        return dataflow.Mux(cond, x, y)
    return x if cond else y

def compare(x, y):
    cond = x < y
    small = mux(cond, x, y)
    large = mux(cond, y, x)
    return small, large

def network(values, num):
    ret = []
    x = values[0]
    
    for i in range(num):
        small, large = compare(x, values[i+1])
        ret.append(small)
        x = large
        
    ret.append(x)
    
    for i in range(num+1, len(values)): 
        ret.append( values[i] )
        
    return ret

def sort(values):
    num = len(values)
    for i in range(num):
        values = network(values, num-i-1)
    return values

def mkSort(numports=8):
    values = [ dataflow.Variable('din%d' % i) for i in range(numports) ]
    rslt = sort(values)

    for i, r in enumerate(rslt):
        r.output('dout%d' % i)

    df = dataflow.Dataflow(*rslt)
    m = df.to_module('sort')

    return m

def mkTest(numports=8):

    m = Module('test')
    
    # target instance
    main = mkSort(numports)

    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    din = [ ports['din%d' % i] for i in range(numports) ]
    dout = [ ports['dout%d' % i] for i in range(numports) ]

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    reset_done = m.Reg('reset_done', initval=0)
    reset_stmt = []
    for i, d in enumerate(din):
        reset_stmt.append( d(0) )
        
    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock

    init.add(
        Delay(1000),
        reset_done(1),
        nclk(clk),
        Delay(100000),
        Systask('finish'),
    )

    fsm = FSM(m, 'fsm', clk, rst)
    
    fsm.goto_next(cond=reset_done)

    for i, d in enumerate(din):
        fsm.add( d(100 - i) )
        
    fsm.goto_next()
    
    for i in range(numports ** 2):
        for d in dout:
            fsm.add( Systask('display', '%s = %d', d.name, d))
        fsm.add( Systask('display', '----') )
        fsm.goto_next()

    fsm.add( Systask('finish') )
    fsm.make_always()

    return m
    
if __name__ == '__main__':
    n = 8
    test = mkTest(n)
    verilog = test.to_verilog('tmp.v')
    #print(verilog)

    # run simulator (Icarus Verilog)
    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    ## only target RTL
    #main = mkSort(n)
    #verilog = main.to_verilog('tmp.v')
    #print(verilog)

    values = [ 100 - i for i in range(n) ]
    rslt = sort(values)
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))))

from veriloggen import *
import veriloggen.types.axi as axi


def mkMain():
    m = Module('main')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    myaxi = axi.AxiMaster(m, 'myaxi', clk, rst)
    myaxi.disable_read()

    fsm = FSM(m, 'fsm', clk, rst)

    # write address
    awaddr = 1024
    awlen = 64

    ack, counter = myaxi.write_request(awaddr, awlen, cond=fsm)
    fsm.If(ack).goto_next()

    # write data
    wdata = m.Reg('wdata', 32, initval=0)

    ack, last = myaxi.write_data(wdata, counter, cond=fsm)

    fsm.If(ack)(
        wdata.inc()
    )
    fsm.If(last).goto_next()

    sum = m.Reg('sum', 32, initval=0)
    expected_sum = (awlen - 1) * awlen // 2

    seq = Seq(m, 'seq', clk, rst)
    seq.If(Ands(myaxi.wdata.wvalid, myaxi.wdata.wready))(
        sum.add(myaxi.wdata.wdata)
    )
    seq.Then().If(myaxi.wdata.wlast).Delay(1)(
        Systask('display', "sum=%d expected_sum=%d", sum, expected_sum)
    )

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

#    # awready (no stall)
#    awready = ports['myaxi_awready']
#    _awready = m.TmpWireLike(awready)
#    _awready.assign(1)
#    m.Always()( awready(_awready) )
#
#    # wready (nostall)
#    wready = ports['myaxi_wready']
#    _wready = m.TmpWireLike(wready)
#    _wready.assign(1)
#    m.Always()( wready(_wready) )

    # awready (with stall)
    waddr_fsm = FSM(m, 'waddr', clk, rst)
    _awlen = m.Reg('_awlen', 32, initval=0)

    waddr_fsm(
        ports['myaxi_awready'](0),
        ports['myaxi_wready'](0),
        _awlen(0)
    )
    waddr_fsm.If(ports['myaxi_awvalid']).goto_next()

    waddr_fsm.If(ports['myaxi_awvalid'])(
        ports['myaxi_awready'](1)
    )
    waddr_fsm.goto_next()

    waddr_fsm(
        ports['myaxi_awready'](0),
        _awlen(ports['myaxi_awlen'])
    )
    waddr_fsm.goto_next()

    # wready (with stall)
    waddr_init = waddr_fsm.current
    waddr_fsm(
        ports['myaxi_wready'](0)
    )
    waddr_fsm.If(ports['myaxi_wvalid']).goto_next()

    waddr_fsm.If(ports['myaxi_wvalid'])(
        ports['myaxi_wready'](1)
    )
    waddr_fsm.goto_next()

    waddr_fsm(
        ports['myaxi_wready'](0),
        _awlen.dec()
    )
    waddr_fsm.goto(waddr_init)
    waddr_fsm.If(_awlen == 0).goto_init()

    # wready (no stall)
#    waddr_fsm(
#        ports['myaxi_wready'](1)
#    )
#    waddr_fsm.Delay(1)(
#        ports['myaxi_wready'](0)
#    )
#    waddr_fsm.If(ports['myaxi_wvalid'])(
#        _awlen.dec()
#    )
#    waddr_fsm.Then().If(_awlen == 0).goto_next()

    # arready (no stall)
    arready = ports['myaxi_arready']
    _arready = m.TmpWireLike(arready)
    _arready.assign(0)
    m.Always()(arready(_arready))

    # rvalid (no stall)
    rvalid = ports['myaxi_rvalid']
    _rvalid = m.TmpWireLike(rvalid)
    _rvalid.assign(0)
    m.Always()(rvalid(_rvalid))

    # rdata (no stall)
    rdata = ports['myaxi_rdata']
    _rdata = m.TmpWireLike(rdata)
    _rdata.assign(0)
    m.Always()(rdata(_rdata))

    # rlast (no stall)
    rlast = ports['myaxi_rlast']
    _rlast = m.TmpWireLike(rlast)
    _rlast.assign(0)
    m.Always()(rlast(_rlast))

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()



from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import subprocess
import tempfile

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module


def setup_waveform(m, *uuts):
    new_uuts = []
    for uut in uuts:
        if isinstance(uut, (tuple, list)):
            for u in uut:
                if isinstance(u, vtypes._Variable) and u.length is not None:
                    continue
                new_uuts.append(u)
        elif isinstance(uut, dict):
            _uut = list(uut.values())
            for u in _uut:
                if isinstance(u, vtypes._Variable) and u.length is not None:
                    continue
                new_uuts.append(u)
        else:
            if isinstance(uut, vtypes._Variable) and uut.length is not None:
                continue
            new_uuts.append(uut)

    uuts = new_uuts
    ret = m.Initial(
        vtypes.Systask('dumpfile', 'uut.vcd'),
        vtypes.Systask('dumpvars', 0, *uuts)
    )
    return ret


def setup_clock(m, clk, hperiod=5):
    ret = m.Initial(
        clk(0),
        vtypes.Forever(clk(vtypes.Not(clk), ldelay=hperiod))
    )
    return ret


def setup_reset(m, reset, *statement, **kwargs):
    period = kwargs['period'] if 'period' in kwargs else 100
    ret = m.Initial(
        reset(0),
        statement,
        vtypes.Delay(100),
        reset(1),
        vtypes.Delay(100),
        reset(0),
    )
    return ret


def next_clock(clk):
    return (vtypes.Event(vtypes.Posedge(clk)), vtypes.Delay(1))


def finish():
    return vtypes.Systask('finish')


class Simulator(object):

    def __init__(self, *objs, **options):
        sim = 'iverilog' if 'sim' not in options else options['sim']
        wave = 'gtkwave' if 'wave' not in options else options['wave']
        files = None if 'files' not in options else options['files']
        top = 'test' if 'top' not in options else options['top']
        self._type_check_sim(sim)
        self._type_check_wave(wave)
        self.objs = objs
        self.files = files
        self.sim = sim
        self.wave = wave
        self.top = top

    def _type_check_sim(self, sim):
        if sim == 'iverilog' or sim == 'icarus':
            return
        if sim == 'modelsim' or sim == 'vsim':
            return
        if sim == 'vcs':
            raise NotImplementedError("Not implemented: '%s'" % sim)
        raise ValueError("Not supported simulator: '%s'" % sim)

    def _type_check_wave(self, wave):
        if wave == 'gtkwave':
            return
        raise ValueError("Not supported waveform viewer: '%s'" % wave)

    def run(self, display=False, outputfile='a.out', include=None, define=None):
        if self.sim == 'iverilog' or self.sim == 'icarus':
            return self._run_iverilog(display, outputfile, include, define)
        if self.sim == 'modelsim' or self.sim == 'vsim':
            return self._run_modelsim(display, self.top, include, define)
        raise NotImplementedError("Not implemented: '%s'" % self.sim)

    def _run_iverilog(self, display=False, outputfile='a.out', include=None, define=None):
        cmd = []
        cmd.append('iverilog')
        if include:
            for inc in include:
                cmd.append('-I')
                cmd.append(inc)
        if define:
            for d in define:
                cmd.append('-D')
                if isinstance(d, (tuple, list)):
                    if d[1] is None:
                        cmd.append(d[0])
                    else:
                        cmd.append(''.join([d[0], '=', str(d[1])]))
                else:
                    cmd.append(d)

        cmd.append('-o')
        cmd.append(outputfile)

        # encoding: 'utf-8' ?
        encode = sys.getdefaultencoding()

        code = self._to_code()
        tmp = tempfile.NamedTemporaryFile()
        tmp.write(code.encode(encode))
        tmp.read()
        filename = tmp.name

        cmd.append(filename)

        # synthesis
        p = subprocess.Popen(' '.join(cmd), shell=True, stdout=subprocess.PIPE)
        syn_rslt = []
        while True:
            stdout_data = p.stdout.readline()
            syn_rslt.append(stdout_data.decode(encode))
            if display:
                print(stdout_data, end='')
            if not stdout_data:
                break
        p.wait()
        p.stdout.close()
        syn_rslt = ''.join(syn_rslt)

        # simulation
        p = subprocess.Popen('./' + outputfile, shell=True,
                             stdout=subprocess.PIPE)
        sim_rslt = []
        while True:
            stdout_data = p.stdout.readline()
            sim_rslt.append(stdout_data.decode(encode))
            if display:
                print(stdout_data, end='')
            if not stdout_data:
                break
        p.wait()
        p.stdout.close()
        sim_rslt = ''.join(sim_rslt)

        # close temporal source code file
        tmp.close()

        return ''.join([syn_rslt, sim_rslt])

    def _run_modelsim(self, display=False, top="test", include=None, define=None):
        cmd = []
        cmd.append('vlib work ; vmap work ; vlog')
        if include:
            for inc in include:
                cmd.append('-I')
                cmd.append(inc)
        if define:
            for d in define:
                cmd.append('-D')
                if isinstance(d, (tuple, list)):
                    if d[1] is None:
                        cmd.append(d[0])
                    else:
                        cmd.append(''.join([d[0], '=', str(d[1])]))
                else:
                    cmd.append(d)

        # encoding: 'utf-8' ?
        encode = sys.getdefaultencoding()

        code = self._to_code()
        tmp = tempfile.NamedTemporaryFile()
        tmp.write(code.encode(encode))
        tmp.read()
        filename = tmp.name

        cmd.append(filename)

        # synthesis
        p = subprocess.Popen(' '.join(cmd), shell=True, stdout=subprocess.PIPE)
        syn_rslt = []
        while True:
            stdout_data = p.stdout.readline()
            syn_rslt.append(stdout_data.decode(encode))
            if display:
                print(stdout_data, end='')
            if not stdout_data:
                break
        p.wait()
        p.stdout.close()
        syn_rslt = ''.join(syn_rslt)

        # simulation
        p = subprocess.Popen(
            'vsim -c ' + top + ' -do \"run -all\"', shell=True, stdout=subprocess.PIPE)
        sim_rslt = []
        while True:
            stdout_data = p.stdout.readline()
            sim_rslt.append(stdout_data.decode(encode))
            if display:
                print(stdout_data, end='')
            if not stdout_data:
                break
        p.wait()
        p.stdout.close()
        sim_rslt = ''.join(sim_rslt)

        # close temporal source code file
        tmp.close()

        return ''.join([syn_rslt, sim_rslt])

    def _to_code(self):
        code = []
        for obj in self.objs:
            if isinstance(obj, module.Module):
                code.append(obj.to_verilog())
                code.append('\n')
            if isinstance(obj, str):
                code.append(obj)
                code.append('\n')
        return ''.join(code)

    def view_waveform(self, filename='uut.vcd', background=False):
        return self._view_waveform_gtkwave(filename, background)

    def _view_waveform_gtkwave(self, filename='uut.vcd', background=False):
        cmd = []
        cmd.append('gtkwave')
        cmd.append('--giga')
        cmd.append(filename)
        if background:
            cmd.append('&')
        subprocess.call(' '.join(cmd), shell=True)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import dataflow_stencil

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg start;
  wire busy;
  reg [4-1:0] ext_src_ram0_addr;
  wire [32-1:0] ext_src_ram0_rdata;
  reg [32-1:0] ext_src_ram0_wdata;
  reg ext_src_ram0_wenable;
  reg [4-1:0] ext_src_ram1_addr;
  wire [32-1:0] ext_src_ram1_rdata;
  reg [32-1:0] ext_src_ram1_wdata;
  reg ext_src_ram1_wenable;
  reg [4-1:0] ext_src_ram2_addr;
  wire [32-1:0] ext_src_ram2_rdata;
  reg [32-1:0] ext_src_ram2_wdata;
  reg ext_src_ram2_wenable;
  reg [4-1:0] ext_dst_ram_addr;
  wire [32-1:0] ext_dst_ram_rdata;
  reg [32-1:0] ext_dst_ram_wdata;
  reg ext_dst_ram_wenable;

  stencil
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .start(start),
    .busy(busy),
    .ext_src_ram0_addr(ext_src_ram0_addr),
    .ext_src_ram0_rdata(ext_src_ram0_rdata),
    .ext_src_ram0_wdata(ext_src_ram0_wdata),
    .ext_src_ram0_wenable(ext_src_ram0_wenable),
    .ext_src_ram1_addr(ext_src_ram1_addr),
    .ext_src_ram1_rdata(ext_src_ram1_rdata),
    .ext_src_ram1_wdata(ext_src_ram1_wdata),
    .ext_src_ram1_wenable(ext_src_ram1_wenable),
    .ext_src_ram2_addr(ext_src_ram2_addr),
    .ext_src_ram2_rdata(ext_src_ram2_rdata),
    .ext_src_ram2_wdata(ext_src_ram2_wdata),
    .ext_src_ram2_wenable(ext_src_ram2_wenable),
    .ext_dst_ram_addr(ext_dst_ram_addr),
    .ext_dst_ram_rdata(ext_dst_ram_rdata),
    .ext_dst_ram_wdata(ext_dst_ram_wdata),
    .ext_dst_ram_wenable(ext_dst_ram_wenable)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    start = 0;
    ext_src_ram0_addr = 0;
    ext_src_ram0_wdata = 0;
    ext_src_ram0_wenable = 0;
    ext_src_ram1_addr = 0;
    ext_src_ram1_wdata = 0;
    ext_src_ram1_wenable = 0;
    ext_src_ram2_addr = 0;
    ext_src_ram2_wdata = 0;
    ext_src_ram2_wenable = 0;
    ext_dst_ram_addr = 2;
    ext_dst_ram_wdata = 0;
    ext_dst_ram_wenable = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #100000;
    $finish;
  end

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] _d1_fsm;
  reg _fsm_cond_4_0_1;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _d1_fsm <= fsm_init;
      _fsm_cond_4_0_1 <= 0;
    end else begin
      _d1_fsm <= fsm;
      case(_d1_fsm)
        fsm_4: begin
          if(_fsm_cond_4_0_1) begin
            start <= 0;
          end 
        end
      endcase
      case(fsm)
        fsm_init: begin
          if(reset_done) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          ext_src_ram0_addr <= -1;
          ext_src_ram1_addr <= -1;
          ext_src_ram2_addr <= -1;
          fsm <= fsm_2;
        end
        fsm_2: begin
          ext_src_ram0_addr <= ext_src_ram0_addr + 1;
          ext_src_ram0_wdata <= 5898240;
          ext_src_ram0_wenable <= 1;
          if(ext_src_ram0_wenable && (ext_src_ram0_addr == 15)) begin
            ext_src_ram0_wenable <= 0;
          end 
          ext_src_ram1_addr <= ext_src_ram1_addr + 1;
          ext_src_ram1_wdata <= 5898240;
          ext_src_ram1_wenable <= 1;
          if(ext_src_ram1_wenable && (ext_src_ram1_addr == 15)) begin
            ext_src_ram1_wenable <= 0;
          end 
          ext_src_ram2_addr <= ext_src_ram2_addr + 1;
          ext_src_ram2_wdata <= 5898240;
          ext_src_ram2_wenable <= 1;
          if(ext_src_ram2_wenable && (ext_src_ram2_addr == 15)) begin
            ext_src_ram2_wenable <= 0;
          end 
          if(ext_src_ram2_wenable && (ext_src_ram0_addr == 15)) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(!busy) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          start <= 1;
          _fsm_cond_4_0_1 <= 1;
          fsm <= fsm_5;
        end
        fsm_5: begin
          if(busy) begin
            fsm <= fsm_6;
          end 
        end
        fsm_6: begin
          if(!busy) begin
            fsm <= fsm_7;
          end 
        end
        fsm_7: begin
          $finish;
        end
      endcase
    end
  end


endmodule



module stencil
(
  input CLK,
  input RST,
  input start,
  output reg busy,
  input [4-1:0] ext_src_ram0_addr,
  output [32-1:0] ext_src_ram0_rdata,
  input [32-1:0] ext_src_ram0_wdata,
  input ext_src_ram0_wenable,
  input [4-1:0] ext_src_ram1_addr,
  output [32-1:0] ext_src_ram1_rdata,
  input [32-1:0] ext_src_ram1_wdata,
  input ext_src_ram1_wenable,
  input [4-1:0] ext_src_ram2_addr,
  output [32-1:0] ext_src_ram2_rdata,
  input [32-1:0] ext_src_ram2_wdata,
  input ext_src_ram2_wenable,
  input [4-1:0] ext_dst_ram_addr,
  output [32-1:0] ext_dst_ram_rdata,
  input [32-1:0] ext_dst_ram_wdata,
  input ext_dst_ram_wenable
);

  reg _tmp_0;
  reg [8-1:0] src_ram0_0_addr;
  wire [32-1:0] src_ram0_0_rdata;
  reg [32-1:0] src_ram0_0_wdata;
  reg src_ram0_0_wenable;
  reg [8-1:0] src_ram0_1_addr;
  wire [32-1:0] src_ram0_1_rdata;
  reg [32-1:0] src_ram0_1_wdata;
  reg src_ram0_1_wenable;

  src_ram0
  inst_src_ram0
  (
    .CLK(CLK),
    .src_ram0_0_addr(src_ram0_0_addr),
    .src_ram0_0_rdata(src_ram0_0_rdata),
    .src_ram0_0_wdata(src_ram0_0_wdata),
    .src_ram0_0_wenable(src_ram0_0_wenable),
    .src_ram0_1_addr(src_ram0_1_addr),
    .src_ram0_1_rdata(src_ram0_1_rdata),
    .src_ram0_1_wdata(src_ram0_1_wdata),
    .src_ram0_1_wenable(src_ram0_1_wenable)
  );

  reg [8-1:0] src_ram1_0_addr;
  wire [32-1:0] src_ram1_0_rdata;
  reg [32-1:0] src_ram1_0_wdata;
  reg src_ram1_0_wenable;
  reg [8-1:0] src_ram1_1_addr;
  wire [32-1:0] src_ram1_1_rdata;
  reg [32-1:0] src_ram1_1_wdata;
  reg src_ram1_1_wenable;

  src_ram1
  inst_src_ram1
  (
    .CLK(CLK),
    .src_ram1_0_addr(src_ram1_0_addr),
    .src_ram1_0_rdata(src_ram1_0_rdata),
    .src_ram1_0_wdata(src_ram1_0_wdata),
    .src_ram1_0_wenable(src_ram1_0_wenable),
    .src_ram1_1_addr(src_ram1_1_addr),
    .src_ram1_1_rdata(src_ram1_1_rdata),
    .src_ram1_1_wdata(src_ram1_1_wdata),
    .src_ram1_1_wenable(src_ram1_1_wenable)
  );

  reg [8-1:0] src_ram2_0_addr;
  wire [32-1:0] src_ram2_0_rdata;
  reg [32-1:0] src_ram2_0_wdata;
  reg src_ram2_0_wenable;
  reg [8-1:0] src_ram2_1_addr;
  wire [32-1:0] src_ram2_1_rdata;
  reg [32-1:0] src_ram2_1_wdata;
  reg src_ram2_1_wenable;

  src_ram2
  inst_src_ram2
  (
    .CLK(CLK),
    .src_ram2_0_addr(src_ram2_0_addr),
    .src_ram2_0_rdata(src_ram2_0_rdata),
    .src_ram2_0_wdata(src_ram2_0_wdata),
    .src_ram2_0_wenable(src_ram2_0_wenable),
    .src_ram2_1_addr(src_ram2_1_addr),
    .src_ram2_1_rdata(src_ram2_1_rdata),
    .src_ram2_1_wdata(src_ram2_1_wdata),
    .src_ram2_1_wenable(src_ram2_1_wenable)
  );

  reg [8-1:0] dst_ram_0_addr;
  wire [32-1:0] dst_ram_0_rdata;
  reg [32-1:0] dst_ram_0_wdata;
  reg dst_ram_0_wenable;
  reg [8-1:0] dst_ram_1_addr;
  wire [32-1:0] dst_ram_1_rdata;
  reg [32-1:0] dst_ram_1_wdata;
  reg dst_ram_1_wenable;

  dst_ram
  inst_dst_ram
  (
    .CLK(CLK),
    .dst_ram_0_addr(dst_ram_0_addr),
    .dst_ram_0_rdata(dst_ram_0_rdata),
    .dst_ram_0_wdata(dst_ram_0_wdata),
    .dst_ram_0_wenable(dst_ram_0_wenable),
    .dst_ram_1_addr(dst_ram_1_addr),
    .dst_ram_1_rdata(dst_ram_1_rdata),
    .dst_ram_1_wdata(dst_ram_1_wdata),
    .dst_ram_1_wenable(dst_ram_1_wenable)
  );

  wire [8-1:0] _tmp_1;
  assign _tmp_1 = ext_src_ram0_addr;

  always @(*) begin
    src_ram0_1_addr = _tmp_1;
  end

  assign ext_src_ram0_rdata = src_ram0_1_rdata;
  wire [32-1:0] _tmp_2;
  assign _tmp_2 = ext_src_ram0_wdata;

  always @(*) begin
    src_ram0_1_wdata = _tmp_2;
  end

  wire _tmp_3;
  assign _tmp_3 = ext_src_ram0_wenable;

  always @(*) begin
    src_ram0_1_wenable = _tmp_3;
  end

  wire [8-1:0] _tmp_4;
  assign _tmp_4 = ext_src_ram1_addr;

  always @(*) begin
    src_ram1_1_addr = _tmp_4;
  end

  assign ext_src_ram1_rdata = src_ram1_1_rdata;
  wire [32-1:0] _tmp_5;
  assign _tmp_5 = ext_src_ram1_wdata;

  always @(*) begin
    src_ram1_1_wdata = _tmp_5;
  end

  wire _tmp_6;
  assign _tmp_6 = ext_src_ram1_wenable;

  always @(*) begin
    src_ram1_1_wenable = _tmp_6;
  end

  wire [8-1:0] _tmp_7;
  assign _tmp_7 = ext_src_ram2_addr;

  always @(*) begin
    src_ram2_1_addr = _tmp_7;
  end

  assign ext_src_ram2_rdata = src_ram2_1_rdata;
  wire [32-1:0] _tmp_8;
  assign _tmp_8 = ext_src_ram2_wdata;

  always @(*) begin
    src_ram2_1_wdata = _tmp_8;
  end

  wire _tmp_9;
  assign _tmp_9 = ext_src_ram2_wenable;

  always @(*) begin
    src_ram2_1_wenable = _tmp_9;
  end

  wire [8-1:0] _tmp_10;
  assign _tmp_10 = ext_dst_ram_addr;

  always @(*) begin
    dst_ram_1_addr = _tmp_10;
  end

  assign ext_dst_ram_rdata = dst_ram_1_rdata;
  wire [32-1:0] _tmp_11;
  assign _tmp_11 = ext_dst_ram_wdata;

  always @(*) begin
    dst_ram_1_wdata = _tmp_11;
  end

  wire _tmp_12;
  assign _tmp_12 = ext_dst_ram_wenable;

  always @(*) begin
    dst_ram_1_wenable = _tmp_12;
  end

  reg [32-1:0] read_fsm;
  localparam read_fsm_init = 0;
  reg [32-1:0] read_count;
  reg [32-1:0] read_addr;
  reg _tmp_13;
  reg _src_ram0_cond_0_1;
  reg _src_ram0_cond_1_1;
  reg _src_ram0_cond_1_2;
  reg _tmp_14;
  reg _src_ram1_cond_0_1;
  reg _src_ram1_cond_1_1;
  reg _src_ram1_cond_1_2;
  reg _tmp_15;
  reg _src_ram2_cond_0_1;
  reg _src_ram2_cond_1_1;
  reg _src_ram2_cond_1_2;
  localparam read_fsm_1 = 1;
  localparam read_fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      read_fsm <= read_fsm_init;
      read_addr <= 0;
      read_count <= 0;
      busy <= 0;
    end else begin
      case(read_fsm)
        read_fsm_init: begin
          read_addr <= 0;
          read_count <= 0;
          busy <= 0;
          if(start) begin
            busy <= 1;
          end 
          if(start) begin
            read_fsm <= read_fsm_1;
          end 
        end
        read_fsm_1: begin
          read_addr <= read_addr + 1;
          read_count <= read_count + 1;
          if(read_count == 15) begin
            read_addr <= 0;
            read_count <= 0;
          end 
          if(read_count == 15) begin
            read_fsm <= read_fsm_2;
          end 
        end
        read_fsm_2: begin
          if(_tmp_0) begin
            busy <= 0;
          end 
          if(_tmp_0) begin
            read_fsm <= read_fsm_init;
          end 
        end
      endcase
    end
  end

  wire [32-1:0] odata;
  wire ovalid;

  stencil_pipeline_2d
  inst_stencil
  (
    .CLK(CLK),
    .RST(RST),
    .idata0(src_ram0_0_rdata),
    .ivalid0(_tmp_13),
    .idata1(src_ram1_0_rdata),
    .ivalid1(_tmp_14),
    .idata2(src_ram2_0_rdata),
    .ivalid2(_tmp_15),
    .odata(odata),
    .ovalid(ovalid)
  );

  reg [32-1:0] write_fsm;
  localparam write_fsm_init = 0;
  reg [32-1:0] write_count;
  reg [32-1:0] write_addr;
  reg _dst_ram_cond_0_1;

  always @(posedge CLK) begin
    if(RST) begin
      write_fsm <= write_fsm_init;
      _tmp_0 <= 0;
      write_addr <= 1;
      write_count <= 0;
    end else begin
      case(write_fsm)
        write_fsm_init: begin
          _tmp_0 <= 0;
          if(ovalid && (write_count > 1)) begin
            write_addr <= write_addr + 1;
          end 
          if(ovalid) begin
            write_count <= write_count + 1;
          end 
          if(write_count == 16) begin
            write_count <= 0;
            write_addr <= 1;
            _tmp_0 <= 1;
          end 
          if(write_count == 16) begin
            write_fsm <= write_fsm_init;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      src_ram0_0_wdata <= 0;
      src_ram0_0_wenable <= 0;
      src_ram0_0_addr <= 0;
      _src_ram0_cond_0_1 <= 0;
      _tmp_13 <= 0;
      _src_ram0_cond_1_1 <= 0;
      _src_ram0_cond_1_2 <= 0;
    end else begin
      if(_src_ram0_cond_1_2) begin
        _tmp_13 <= 0;
      end 
      if(_src_ram0_cond_0_1) begin
        _tmp_13 <= 1;
      end 
      _src_ram0_cond_1_2 <= _src_ram0_cond_1_1;
      src_ram0_0_wdata <= 0;
      src_ram0_0_wenable <= 0;
      if(read_fsm == 1) begin
        src_ram0_0_addr <= read_addr;
      end 
      _src_ram0_cond_0_1 <= read_fsm == 1;
      _src_ram0_cond_1_1 <= read_fsm == 1;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      src_ram1_0_wdata <= 0;
      src_ram1_0_wenable <= 0;
      src_ram1_0_addr <= 0;
      _src_ram1_cond_0_1 <= 0;
      _tmp_14 <= 0;
      _src_ram1_cond_1_1 <= 0;
      _src_ram1_cond_1_2 <= 0;
    end else begin
      if(_src_ram1_cond_1_2) begin
        _tmp_14 <= 0;
      end 
      if(_src_ram1_cond_0_1) begin
        _tmp_14 <= 1;
      end 
      _src_ram1_cond_1_2 <= _src_ram1_cond_1_1;
      src_ram1_0_wdata <= 0;
      src_ram1_0_wenable <= 0;
      if(read_fsm == 1) begin
        src_ram1_0_addr <= read_addr;
      end 
      _src_ram1_cond_0_1 <= read_fsm == 1;
      _src_ram1_cond_1_1 <= read_fsm == 1;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      src_ram2_0_wdata <= 0;
      src_ram2_0_wenable <= 0;
      src_ram2_0_addr <= 0;
      _src_ram2_cond_0_1 <= 0;
      _tmp_15 <= 0;
      _src_ram2_cond_1_1 <= 0;
      _src_ram2_cond_1_2 <= 0;
    end else begin
      if(_src_ram2_cond_1_2) begin
        _tmp_15 <= 0;
      end 
      if(_src_ram2_cond_0_1) begin
        _tmp_15 <= 1;
      end 
      _src_ram2_cond_1_2 <= _src_ram2_cond_1_1;
      src_ram2_0_wdata <= 0;
      src_ram2_0_wenable <= 0;
      if(read_fsm == 1) begin
        src_ram2_0_addr <= read_addr;
      end 
      _src_ram2_cond_0_1 <= read_fsm == 1;
      _src_ram2_cond_1_1 <= read_fsm == 1;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      dst_ram_0_addr <= 0;
      dst_ram_0_wdata <= 0;
      dst_ram_0_wenable <= 0;
      _dst_ram_cond_0_1 <= 0;
    end else begin
      if(_dst_ram_cond_0_1) begin
        dst_ram_0_wenable <= 0;
      end 
      if((write_fsm == 0) && (ovalid && (write_count > 1))) begin
        dst_ram_0_addr <= write_addr;
        dst_ram_0_wdata <= odata;
        dst_ram_0_wenable <= 1;
      end 
      _dst_ram_cond_0_1 <= (write_fsm == 0) && (ovalid && (write_count > 1));
    end
  end


endmodule



module src_ram0
(
  input CLK,
  input [8-1:0] src_ram0_0_addr,
  output [32-1:0] src_ram0_0_rdata,
  input [32-1:0] src_ram0_0_wdata,
  input src_ram0_0_wenable,
  input [8-1:0] src_ram0_1_addr,
  output [32-1:0] src_ram0_1_rdata,
  input [32-1:0] src_ram0_1_wdata,
  input src_ram0_1_wenable
);

  reg [8-1:0] src_ram0_0_daddr;
  reg [8-1:0] src_ram0_1_daddr;
  reg [32-1:0] mem [0:256-1];

  always @(posedge CLK) begin
    if(src_ram0_0_wenable) begin
      mem[src_ram0_0_addr] <= src_ram0_0_wdata;
    end 
    src_ram0_0_daddr <= src_ram0_0_addr;
  end

  assign src_ram0_0_rdata = mem[src_ram0_0_daddr];

  always @(posedge CLK) begin
    if(src_ram0_1_wenable) begin
      mem[src_ram0_1_addr] <= src_ram0_1_wdata;
    end 
    src_ram0_1_daddr <= src_ram0_1_addr;
  end

  assign src_ram0_1_rdata = mem[src_ram0_1_daddr];

endmodule



module src_ram1
(
  input CLK,
  input [8-1:0] src_ram1_0_addr,
  output [32-1:0] src_ram1_0_rdata,
  input [32-1:0] src_ram1_0_wdata,
  input src_ram1_0_wenable,
  input [8-1:0] src_ram1_1_addr,
  output [32-1:0] src_ram1_1_rdata,
  input [32-1:0] src_ram1_1_wdata,
  input src_ram1_1_wenable
);

  reg [8-1:0] src_ram1_0_daddr;
  reg [8-1:0] src_ram1_1_daddr;
  reg [32-1:0] mem [0:256-1];

  always @(posedge CLK) begin
    if(src_ram1_0_wenable) begin
      mem[src_ram1_0_addr] <= src_ram1_0_wdata;
    end 
    src_ram1_0_daddr <= src_ram1_0_addr;
  end

  assign src_ram1_0_rdata = mem[src_ram1_0_daddr];

  always @(posedge CLK) begin
    if(src_ram1_1_wenable) begin
      mem[src_ram1_1_addr] <= src_ram1_1_wdata;
    end 
    src_ram1_1_daddr <= src_ram1_1_addr;
  end

  assign src_ram1_1_rdata = mem[src_ram1_1_daddr];

endmodule



module src_ram2
(
  input CLK,
  input [8-1:0] src_ram2_0_addr,
  output [32-1:0] src_ram2_0_rdata,
  input [32-1:0] src_ram2_0_wdata,
  input src_ram2_0_wenable,
  input [8-1:0] src_ram2_1_addr,
  output [32-1:0] src_ram2_1_rdata,
  input [32-1:0] src_ram2_1_wdata,
  input src_ram2_1_wenable
);

  reg [8-1:0] src_ram2_0_daddr;
  reg [8-1:0] src_ram2_1_daddr;
  reg [32-1:0] mem [0:256-1];

  always @(posedge CLK) begin
    if(src_ram2_0_wenable) begin
      mem[src_ram2_0_addr] <= src_ram2_0_wdata;
    end 
    src_ram2_0_daddr <= src_ram2_0_addr;
  end

  assign src_ram2_0_rdata = mem[src_ram2_0_daddr];

  always @(posedge CLK) begin
    if(src_ram2_1_wenable) begin
      mem[src_ram2_1_addr] <= src_ram2_1_wdata;
    end 
    src_ram2_1_daddr <= src_ram2_1_addr;
  end

  assign src_ram2_1_rdata = mem[src_ram2_1_daddr];

endmodule



module dst_ram
(
  input CLK,
  input [8-1:0] dst_ram_0_addr,
  output [32-1:0] dst_ram_0_rdata,
  input [32-1:0] dst_ram_0_wdata,
  input dst_ram_0_wenable,
  input [8-1:0] dst_ram_1_addr,
  output [32-1:0] dst_ram_1_rdata,
  input [32-1:0] dst_ram_1_wdata,
  input dst_ram_1_wenable
);

  reg [8-1:0] dst_ram_0_daddr;
  reg [8-1:0] dst_ram_1_daddr;
  reg [32-1:0] mem [0:256-1];

  always @(posedge CLK) begin
    if(dst_ram_0_wenable) begin
      mem[dst_ram_0_addr] <= dst_ram_0_wdata;
    end 
    dst_ram_0_daddr <= dst_ram_0_addr;
  end

  assign dst_ram_0_rdata = mem[dst_ram_0_daddr];

  always @(posedge CLK) begin
    if(dst_ram_1_wenable) begin
      mem[dst_ram_1_addr] <= dst_ram_1_wdata;
    end 
    dst_ram_1_daddr <= dst_ram_1_addr;
  end

  assign dst_ram_1_rdata = mem[dst_ram_1_daddr];

endmodule



module stencil_pipeline_2d
(
  input CLK,
  input RST,
  input signed [32-1:0] idata0,
  input ivalid0,
  input signed [32-1:0] idata1,
  input ivalid1,
  input signed [32-1:0] idata2,
  input ivalid2,
  output signed [32-1:0] odata,
  output ovalid
);

  reg signed [32-1:0] _tmp_data_0;
  reg signed [32-1:0] _tmp_data_1;
  reg signed [32-1:0] _tmp_data_2;
  reg signed [32-1:0] _tmp_data_3;
  reg signed [32-1:0] _tmp_data_4;
  reg signed [32-1:0] _tmp_data_5;
  wire signed [32-1:0] _tmp_data_6;
  wire _tmp_valid_6;
  wire _tmp_ready_6;
  wire signed [46-1:0] _tmp_odata_6;
  reg signed [46-1:0] _tmp_data_reg_6;
  assign _tmp_data_6 = _tmp_data_reg_6;
  wire _tmp_ovalid_6;
  reg _tmp_valid_reg_6;
  assign _tmp_valid_6 = _tmp_valid_reg_6;
  wire _tmp_enable_6;
  wire _tmp_update_6;
  assign _tmp_enable_6 = (_tmp_ready_6 || !_tmp_valid_6) && 1 && ivalid0;
  assign _tmp_update_6 = _tmp_ready_6 || !_tmp_valid_6;

  multiplier_0
  mul6
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_6),
    .enable(_tmp_enable_6),
    .valid(_tmp_ovalid_6),
    .a(idata0),
    .b(14'd7281),
    .c(_tmp_odata_6)
  );

  assign _tmp_ready_6 = (_tmp_ready_15 || !_tmp_valid_15) && (_tmp_valid_6 && _tmp_valid_7);
  wire signed [32-1:0] _tmp_data_7;
  wire _tmp_valid_7;
  wire _tmp_ready_7;
  wire signed [46-1:0] _tmp_odata_7;
  reg signed [46-1:0] _tmp_data_reg_7;
  assign _tmp_data_7 = _tmp_data_reg_7;
  wire _tmp_ovalid_7;
  reg _tmp_valid_reg_7;
  assign _tmp_valid_7 = _tmp_valid_reg_7;
  wire _tmp_enable_7;
  wire _tmp_update_7;
  assign _tmp_enable_7 = (_tmp_ready_7 || !_tmp_valid_7) && 1 && ivalid0;
  assign _tmp_update_7 = _tmp_ready_7 || !_tmp_valid_7;

  multiplier_1
  mul7
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_7),
    .enable(_tmp_enable_7),
    .valid(_tmp_ovalid_7),
    .a(_tmp_data_0),
    .b(14'd7281),
    .c(_tmp_odata_7)
  );

  assign _tmp_ready_7 = (_tmp_ready_15 || !_tmp_valid_15) && (_tmp_valid_6 && _tmp_valid_7);
  wire signed [32-1:0] _tmp_data_8;
  wire _tmp_valid_8;
  wire _tmp_ready_8;
  wire signed [46-1:0] _tmp_odata_8;
  reg signed [46-1:0] _tmp_data_reg_8;
  assign _tmp_data_8 = _tmp_data_reg_8;
  wire _tmp_ovalid_8;
  reg _tmp_valid_reg_8;
  assign _tmp_valid_8 = _tmp_valid_reg_8;
  wire _tmp_enable_8;
  wire _tmp_update_8;
  assign _tmp_enable_8 = (_tmp_ready_8 || !_tmp_valid_8) && 1 && ivalid0;
  assign _tmp_update_8 = _tmp_ready_8 || !_tmp_valid_8;

  multiplier_2
  mul8
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_8),
    .enable(_tmp_enable_8),
    .valid(_tmp_ovalid_8),
    .a(_tmp_data_1),
    .b(14'd7281),
    .c(_tmp_odata_8)
  );

  assign _tmp_ready_8 = (_tmp_ready_16 || !_tmp_valid_16) && _tmp_valid_8;
  wire signed [32-1:0] _tmp_data_9;
  wire _tmp_valid_9;
  wire _tmp_ready_9;
  wire signed [46-1:0] _tmp_odata_9;
  reg signed [46-1:0] _tmp_data_reg_9;
  assign _tmp_data_9 = _tmp_data_reg_9;
  wire _tmp_ovalid_9;
  reg _tmp_valid_reg_9;
  assign _tmp_valid_9 = _tmp_valid_reg_9;
  wire _tmp_enable_9;
  wire _tmp_update_9;
  assign _tmp_enable_9 = (_tmp_ready_9 || !_tmp_valid_9) && 1 && ivalid1;
  assign _tmp_update_9 = _tmp_ready_9 || !_tmp_valid_9;

  multiplier_3
  mul9
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_9),
    .enable(_tmp_enable_9),
    .valid(_tmp_ovalid_9),
    .a(idata1),
    .b(14'd7281),
    .c(_tmp_odata_9)
  );

  assign _tmp_ready_9 = (_tmp_ready_17 || !_tmp_valid_17) && _tmp_valid_9;
  wire signed [32-1:0] _tmp_data_10;
  wire _tmp_valid_10;
  wire _tmp_ready_10;
  wire signed [46-1:0] _tmp_odata_10;
  reg signed [46-1:0] _tmp_data_reg_10;
  assign _tmp_data_10 = _tmp_data_reg_10;
  wire _tmp_ovalid_10;
  reg _tmp_valid_reg_10;
  assign _tmp_valid_10 = _tmp_valid_reg_10;
  wire _tmp_enable_10;
  wire _tmp_update_10;
  assign _tmp_enable_10 = (_tmp_ready_10 || !_tmp_valid_10) && 1 && ivalid1;
  assign _tmp_update_10 = _tmp_ready_10 || !_tmp_valid_10;

  multiplier_4
  mul10
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_10),
    .enable(_tmp_enable_10),
    .valid(_tmp_ovalid_10),
    .a(_tmp_data_2),
    .b(14'd7281),
    .c(_tmp_odata_10)
  );

  assign _tmp_ready_10 = (_tmp_ready_18 || !_tmp_valid_18) && _tmp_valid_10;
  wire signed [32-1:0] _tmp_data_11;
  wire _tmp_valid_11;
  wire _tmp_ready_11;
  wire signed [46-1:0] _tmp_odata_11;
  reg signed [46-1:0] _tmp_data_reg_11;
  assign _tmp_data_11 = _tmp_data_reg_11;
  wire _tmp_ovalid_11;
  reg _tmp_valid_reg_11;
  assign _tmp_valid_11 = _tmp_valid_reg_11;
  wire _tmp_enable_11;
  wire _tmp_update_11;
  assign _tmp_enable_11 = (_tmp_ready_11 || !_tmp_valid_11) && 1 && ivalid1;
  assign _tmp_update_11 = _tmp_ready_11 || !_tmp_valid_11;

  multiplier_5
  mul11
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_11),
    .enable(_tmp_enable_11),
    .valid(_tmp_ovalid_11),
    .a(_tmp_data_3),
    .b(14'd7281),
    .c(_tmp_odata_11)
  );

  assign _tmp_ready_11 = (_tmp_ready_19 || !_tmp_valid_19) && _tmp_valid_11;
  wire signed [32-1:0] _tmp_data_12;
  wire _tmp_valid_12;
  wire _tmp_ready_12;
  wire signed [46-1:0] _tmp_odata_12;
  reg signed [46-1:0] _tmp_data_reg_12;
  assign _tmp_data_12 = _tmp_data_reg_12;
  wire _tmp_ovalid_12;
  reg _tmp_valid_reg_12;
  assign _tmp_valid_12 = _tmp_valid_reg_12;
  wire _tmp_enable_12;
  wire _tmp_update_12;
  assign _tmp_enable_12 = (_tmp_ready_12 || !_tmp_valid_12) && 1 && ivalid2;
  assign _tmp_update_12 = _tmp_ready_12 || !_tmp_valid_12;

  multiplier_6
  mul12
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_12),
    .enable(_tmp_enable_12),
    .valid(_tmp_ovalid_12),
    .a(idata2),
    .b(14'd7281),
    .c(_tmp_odata_12)
  );

  assign _tmp_ready_12 = (_tmp_ready_20 || !_tmp_valid_20) && _tmp_valid_12;
  wire signed [32-1:0] _tmp_data_13;
  wire _tmp_valid_13;
  wire _tmp_ready_13;
  wire signed [46-1:0] _tmp_odata_13;
  reg signed [46-1:0] _tmp_data_reg_13;
  assign _tmp_data_13 = _tmp_data_reg_13;
  wire _tmp_ovalid_13;
  reg _tmp_valid_reg_13;
  assign _tmp_valid_13 = _tmp_valid_reg_13;
  wire _tmp_enable_13;
  wire _tmp_update_13;
  assign _tmp_enable_13 = (_tmp_ready_13 || !_tmp_valid_13) && 1 && ivalid2;
  assign _tmp_update_13 = _tmp_ready_13 || !_tmp_valid_13;

  multiplier_7
  mul13
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_13),
    .enable(_tmp_enable_13),
    .valid(_tmp_ovalid_13),
    .a(_tmp_data_4),
    .b(14'd7281),
    .c(_tmp_odata_13)
  );

  assign _tmp_ready_13 = (_tmp_ready_21 || !_tmp_valid_21) && _tmp_valid_13;
  wire signed [32-1:0] _tmp_data_14;
  wire _tmp_valid_14;
  wire _tmp_ready_14;
  wire signed [46-1:0] _tmp_odata_14;
  reg signed [46-1:0] _tmp_data_reg_14;
  assign _tmp_data_14 = _tmp_data_reg_14;
  wire _tmp_ovalid_14;
  reg _tmp_valid_reg_14;
  assign _tmp_valid_14 = _tmp_valid_reg_14;
  wire _tmp_enable_14;
  wire _tmp_update_14;
  assign _tmp_enable_14 = (_tmp_ready_14 || !_tmp_valid_14) && 1 && ivalid2;
  assign _tmp_update_14 = _tmp_ready_14 || !_tmp_valid_14;

  multiplier_8
  mul14
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_14),
    .enable(_tmp_enable_14),
    .valid(_tmp_ovalid_14),
    .a(_tmp_data_5),
    .b(14'd7281),
    .c(_tmp_odata_14)
  );

  assign _tmp_ready_14 = (_tmp_ready_22 || !_tmp_valid_22) && _tmp_valid_14;
  reg signed [32-1:0] _tmp_data_15;
  reg _tmp_valid_15;
  wire _tmp_ready_15;
  assign _tmp_ready_15 = (_tmp_ready_23 || !_tmp_valid_23) && (_tmp_valid_15 && _tmp_valid_16);
  reg signed [32-1:0] _tmp_data_16;
  reg _tmp_valid_16;
  wire _tmp_ready_16;
  assign _tmp_ready_16 = (_tmp_ready_23 || !_tmp_valid_23) && (_tmp_valid_15 && _tmp_valid_16);
  reg signed [32-1:0] _tmp_data_17;
  reg _tmp_valid_17;
  wire _tmp_ready_17;
  assign _tmp_ready_17 = (_tmp_ready_24 || !_tmp_valid_24) && _tmp_valid_17;
  reg signed [32-1:0] _tmp_data_18;
  reg _tmp_valid_18;
  wire _tmp_ready_18;
  assign _tmp_ready_18 = (_tmp_ready_25 || !_tmp_valid_25) && _tmp_valid_18;
  reg signed [32-1:0] _tmp_data_19;
  reg _tmp_valid_19;
  wire _tmp_ready_19;
  assign _tmp_ready_19 = (_tmp_ready_26 || !_tmp_valid_26) && _tmp_valid_19;
  reg signed [32-1:0] _tmp_data_20;
  reg _tmp_valid_20;
  wire _tmp_ready_20;
  assign _tmp_ready_20 = (_tmp_ready_27 || !_tmp_valid_27) && _tmp_valid_20;
  reg signed [32-1:0] _tmp_data_21;
  reg _tmp_valid_21;
  wire _tmp_ready_21;
  assign _tmp_ready_21 = (_tmp_ready_28 || !_tmp_valid_28) && _tmp_valid_21;
  reg signed [32-1:0] _tmp_data_22;
  reg _tmp_valid_22;
  wire _tmp_ready_22;
  assign _tmp_ready_22 = (_tmp_ready_29 || !_tmp_valid_29) && _tmp_valid_22;
  reg signed [32-1:0] _tmp_data_23;
  reg _tmp_valid_23;
  wire _tmp_ready_23;
  assign _tmp_ready_23 = (_tmp_ready_30 || !_tmp_valid_30) && (_tmp_valid_23 && _tmp_valid_24);
  reg signed [32-1:0] _tmp_data_24;
  reg _tmp_valid_24;
  wire _tmp_ready_24;
  assign _tmp_ready_24 = (_tmp_ready_30 || !_tmp_valid_30) && (_tmp_valid_23 && _tmp_valid_24);
  reg signed [32-1:0] _tmp_data_25;
  reg _tmp_valid_25;
  wire _tmp_ready_25;
  assign _tmp_ready_25 = (_tmp_ready_31 || !_tmp_valid_31) && _tmp_valid_25;
  reg signed [32-1:0] _tmp_data_26;
  reg _tmp_valid_26;
  wire _tmp_ready_26;
  assign _tmp_ready_26 = (_tmp_ready_32 || !_tmp_valid_32) && _tmp_valid_26;
  reg signed [32-1:0] _tmp_data_27;
  reg _tmp_valid_27;
  wire _tmp_ready_27;
  assign _tmp_ready_27 = (_tmp_ready_33 || !_tmp_valid_33) && _tmp_valid_27;
  reg signed [32-1:0] _tmp_data_28;
  reg _tmp_valid_28;
  wire _tmp_ready_28;
  assign _tmp_ready_28 = (_tmp_ready_34 || !_tmp_valid_34) && _tmp_valid_28;
  reg signed [32-1:0] _tmp_data_29;
  reg _tmp_valid_29;
  wire _tmp_ready_29;
  assign _tmp_ready_29 = (_tmp_ready_35 || !_tmp_valid_35) && _tmp_valid_29;
  reg signed [32-1:0] _tmp_data_30;
  reg _tmp_valid_30;
  wire _tmp_ready_30;
  assign _tmp_ready_30 = (_tmp_ready_36 || !_tmp_valid_36) && (_tmp_valid_30 && _tmp_valid_31);
  reg signed [32-1:0] _tmp_data_31;
  reg _tmp_valid_31;
  wire _tmp_ready_31;
  assign _tmp_ready_31 = (_tmp_ready_36 || !_tmp_valid_36) && (_tmp_valid_30 && _tmp_valid_31);
  reg signed [32-1:0] _tmp_data_32;
  reg _tmp_valid_32;
  wire _tmp_ready_32;
  assign _tmp_ready_32 = (_tmp_ready_37 || !_tmp_valid_37) && _tmp_valid_32;
  reg signed [32-1:0] _tmp_data_33;
  reg _tmp_valid_33;
  wire _tmp_ready_33;
  assign _tmp_ready_33 = (_tmp_ready_38 || !_tmp_valid_38) && _tmp_valid_33;
  reg signed [32-1:0] _tmp_data_34;
  reg _tmp_valid_34;
  wire _tmp_ready_34;
  assign _tmp_ready_34 = (_tmp_ready_39 || !_tmp_valid_39) && _tmp_valid_34;
  reg signed [32-1:0] _tmp_data_35;
  reg _tmp_valid_35;
  wire _tmp_ready_35;
  assign _tmp_ready_35 = (_tmp_ready_40 || !_tmp_valid_40) && _tmp_valid_35;
  reg signed [32-1:0] _tmp_data_36;
  reg _tmp_valid_36;
  wire _tmp_ready_36;
  assign _tmp_ready_36 = (_tmp_ready_41 || !_tmp_valid_41) && (_tmp_valid_36 && _tmp_valid_37);
  reg signed [32-1:0] _tmp_data_37;
  reg _tmp_valid_37;
  wire _tmp_ready_37;
  assign _tmp_ready_37 = (_tmp_ready_41 || !_tmp_valid_41) && (_tmp_valid_36 && _tmp_valid_37);
  reg signed [32-1:0] _tmp_data_38;
  reg _tmp_valid_38;
  wire _tmp_ready_38;
  assign _tmp_ready_38 = (_tmp_ready_42 || !_tmp_valid_42) && _tmp_valid_38;
  reg signed [32-1:0] _tmp_data_39;
  reg _tmp_valid_39;
  wire _tmp_ready_39;
  assign _tmp_ready_39 = (_tmp_ready_43 || !_tmp_valid_43) && _tmp_valid_39;
  reg signed [32-1:0] _tmp_data_40;
  reg _tmp_valid_40;
  wire _tmp_ready_40;
  assign _tmp_ready_40 = (_tmp_ready_44 || !_tmp_valid_44) && _tmp_valid_40;
  reg signed [32-1:0] _tmp_data_41;
  reg _tmp_valid_41;
  wire _tmp_ready_41;
  assign _tmp_ready_41 = (_tmp_ready_45 || !_tmp_valid_45) && (_tmp_valid_41 && _tmp_valid_42);
  reg signed [32-1:0] _tmp_data_42;
  reg _tmp_valid_42;
  wire _tmp_ready_42;
  assign _tmp_ready_42 = (_tmp_ready_45 || !_tmp_valid_45) && (_tmp_valid_41 && _tmp_valid_42);
  reg signed [32-1:0] _tmp_data_43;
  reg _tmp_valid_43;
  wire _tmp_ready_43;
  assign _tmp_ready_43 = (_tmp_ready_46 || !_tmp_valid_46) && _tmp_valid_43;
  reg signed [32-1:0] _tmp_data_44;
  reg _tmp_valid_44;
  wire _tmp_ready_44;
  assign _tmp_ready_44 = (_tmp_ready_47 || !_tmp_valid_47) && _tmp_valid_44;
  reg signed [32-1:0] _tmp_data_45;
  reg _tmp_valid_45;
  wire _tmp_ready_45;
  assign _tmp_ready_45 = (_tmp_ready_48 || !_tmp_valid_48) && (_tmp_valid_45 && _tmp_valid_46);
  reg signed [32-1:0] _tmp_data_46;
  reg _tmp_valid_46;
  wire _tmp_ready_46;
  assign _tmp_ready_46 = (_tmp_ready_48 || !_tmp_valid_48) && (_tmp_valid_45 && _tmp_valid_46);
  reg signed [32-1:0] _tmp_data_47;
  reg _tmp_valid_47;
  wire _tmp_ready_47;
  assign _tmp_ready_47 = (_tmp_ready_49 || !_tmp_valid_49) && _tmp_valid_47;
  reg signed [32-1:0] _tmp_data_48;
  reg _tmp_valid_48;
  wire _tmp_ready_48;
  assign _tmp_ready_48 = (_tmp_ready_50 || !_tmp_valid_50) && (_tmp_valid_48 && _tmp_valid_49);
  reg signed [32-1:0] _tmp_data_49;
  reg _tmp_valid_49;
  wire _tmp_ready_49;
  assign _tmp_ready_49 = (_tmp_ready_50 || !_tmp_valid_50) && (_tmp_valid_48 && _tmp_valid_49);
  reg signed [32-1:0] _tmp_data_50;
  reg _tmp_valid_50;
  wire _tmp_ready_50;
  assign odata = _tmp_data_50;
  assign ovalid = _tmp_valid_50;
  assign _tmp_ready_50 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_0 <= 0;
      _tmp_data_1 <= 0;
      _tmp_data_2 <= 0;
      _tmp_data_3 <= 0;
      _tmp_data_4 <= 0;
      _tmp_data_5 <= 0;
      _tmp_data_reg_6 <= 0;
      _tmp_valid_reg_6 <= 0;
      _tmp_data_reg_7 <= 0;
      _tmp_valid_reg_7 <= 0;
      _tmp_data_reg_8 <= 0;
      _tmp_valid_reg_8 <= 0;
      _tmp_data_reg_9 <= 0;
      _tmp_valid_reg_9 <= 0;
      _tmp_data_reg_10 <= 0;
      _tmp_valid_reg_10 <= 0;
      _tmp_data_reg_11 <= 0;
      _tmp_valid_reg_11 <= 0;
      _tmp_data_reg_12 <= 0;
      _tmp_valid_reg_12 <= 0;
      _tmp_data_reg_13 <= 0;
      _tmp_valid_reg_13 <= 0;
      _tmp_data_reg_14 <= 0;
      _tmp_valid_reg_14 <= 0;
      _tmp_data_15 <= 0;
      _tmp_valid_15 <= 0;
      _tmp_data_16 <= 0;
      _tmp_valid_16 <= 0;
      _tmp_data_17 <= 0;
      _tmp_valid_17 <= 0;
      _tmp_data_18 <= 0;
      _tmp_valid_18 <= 0;
      _tmp_data_19 <= 0;
      _tmp_valid_19 <= 0;
      _tmp_data_20 <= 0;
      _tmp_valid_20 <= 0;
      _tmp_data_21 <= 0;
      _tmp_valid_21 <= 0;
      _tmp_data_22 <= 0;
      _tmp_valid_22 <= 0;
      _tmp_data_23 <= 0;
      _tmp_valid_23 <= 0;
      _tmp_data_24 <= 0;
      _tmp_valid_24 <= 0;
      _tmp_data_25 <= 0;
      _tmp_valid_25 <= 0;
      _tmp_data_26 <= 0;
      _tmp_valid_26 <= 0;
      _tmp_data_27 <= 0;
      _tmp_valid_27 <= 0;
      _tmp_data_28 <= 0;
      _tmp_valid_28 <= 0;
      _tmp_data_29 <= 0;
      _tmp_valid_29 <= 0;
      _tmp_data_30 <= 0;
      _tmp_valid_30 <= 0;
      _tmp_data_31 <= 0;
      _tmp_valid_31 <= 0;
      _tmp_data_32 <= 0;
      _tmp_valid_32 <= 0;
      _tmp_data_33 <= 0;
      _tmp_valid_33 <= 0;
      _tmp_data_34 <= 0;
      _tmp_valid_34 <= 0;
      _tmp_data_35 <= 0;
      _tmp_valid_35 <= 0;
      _tmp_data_36 <= 0;
      _tmp_valid_36 <= 0;
      _tmp_data_37 <= 0;
      _tmp_valid_37 <= 0;
      _tmp_data_38 <= 0;
      _tmp_valid_38 <= 0;
      _tmp_data_39 <= 0;
      _tmp_valid_39 <= 0;
      _tmp_data_40 <= 0;
      _tmp_valid_40 <= 0;
      _tmp_data_41 <= 0;
      _tmp_valid_41 <= 0;
      _tmp_data_42 <= 0;
      _tmp_valid_42 <= 0;
      _tmp_data_43 <= 0;
      _tmp_valid_43 <= 0;
      _tmp_data_44 <= 0;
      _tmp_valid_44 <= 0;
      _tmp_data_45 <= 0;
      _tmp_valid_45 <= 0;
      _tmp_data_46 <= 0;
      _tmp_valid_46 <= 0;
      _tmp_data_47 <= 0;
      _tmp_valid_47 <= 0;
      _tmp_data_48 <= 0;
      _tmp_valid_48 <= 0;
      _tmp_data_49 <= 0;
      _tmp_valid_49 <= 0;
      _tmp_data_50 <= 0;
      _tmp_valid_50 <= 0;
    end else begin
      if(ivalid0) begin
        _tmp_data_0 <= idata0;
      end 
      if(ivalid0) begin
        _tmp_data_1 <= _tmp_data_0;
      end 
      if(ivalid1) begin
        _tmp_data_2 <= idata1;
      end 
      if(ivalid1) begin
        _tmp_data_3 <= _tmp_data_2;
      end 
      if(ivalid2) begin
        _tmp_data_4 <= idata2;
      end 
      if(ivalid2) begin
        _tmp_data_5 <= _tmp_data_4;
      end 
      if(_tmp_ready_6 || !_tmp_valid_6) begin
        _tmp_data_reg_6 <= _tmp_odata_6 >>> 16;
      end 
      if(_tmp_ready_6 || !_tmp_valid_6) begin
        _tmp_valid_reg_6 <= _tmp_ovalid_6;
      end 
      if(_tmp_ready_7 || !_tmp_valid_7) begin
        _tmp_data_reg_7 <= _tmp_odata_7 >>> 16;
      end 
      if(_tmp_ready_7 || !_tmp_valid_7) begin
        _tmp_valid_reg_7 <= _tmp_ovalid_7;
      end 
      if(_tmp_ready_8 || !_tmp_valid_8) begin
        _tmp_data_reg_8 <= _tmp_odata_8 >>> 16;
      end 
      if(_tmp_ready_8 || !_tmp_valid_8) begin
        _tmp_valid_reg_8 <= _tmp_ovalid_8;
      end 
      if(_tmp_ready_9 || !_tmp_valid_9) begin
        _tmp_data_reg_9 <= _tmp_odata_9 >>> 16;
      end 
      if(_tmp_ready_9 || !_tmp_valid_9) begin
        _tmp_valid_reg_9 <= _tmp_ovalid_9;
      end 
      if(_tmp_ready_10 || !_tmp_valid_10) begin
        _tmp_data_reg_10 <= _tmp_odata_10 >>> 16;
      end 
      if(_tmp_ready_10 || !_tmp_valid_10) begin
        _tmp_valid_reg_10 <= _tmp_ovalid_10;
      end 
      if(_tmp_ready_11 || !_tmp_valid_11) begin
        _tmp_data_reg_11 <= _tmp_odata_11 >>> 16;
      end 
      if(_tmp_ready_11 || !_tmp_valid_11) begin
        _tmp_valid_reg_11 <= _tmp_ovalid_11;
      end 
      if(_tmp_ready_12 || !_tmp_valid_12) begin
        _tmp_data_reg_12 <= _tmp_odata_12 >>> 16;
      end 
      if(_tmp_ready_12 || !_tmp_valid_12) begin
        _tmp_valid_reg_12 <= _tmp_ovalid_12;
      end 
      if(_tmp_ready_13 || !_tmp_valid_13) begin
        _tmp_data_reg_13 <= _tmp_odata_13 >>> 16;
      end 
      if(_tmp_ready_13 || !_tmp_valid_13) begin
        _tmp_valid_reg_13 <= _tmp_ovalid_13;
      end 
      if(_tmp_ready_14 || !_tmp_valid_14) begin
        _tmp_data_reg_14 <= _tmp_odata_14 >>> 16;
      end 
      if(_tmp_ready_14 || !_tmp_valid_14) begin
        _tmp_valid_reg_14 <= _tmp_ovalid_14;
      end 
      if((_tmp_ready_15 || !_tmp_valid_15) && (_tmp_ready_6 && _tmp_ready_7) && (_tmp_valid_6 && _tmp_valid_7)) begin
        _tmp_data_15 <= $signed(_tmp_data_6) + $signed(_tmp_data_7);
      end 
      if(_tmp_valid_15 && _tmp_ready_15) begin
        _tmp_valid_15 <= 0;
      end 
      if((_tmp_ready_15 || !_tmp_valid_15) && (_tmp_ready_6 && _tmp_ready_7)) begin
        _tmp_valid_15 <= _tmp_valid_6 && _tmp_valid_7;
      end 
      if((_tmp_ready_16 || !_tmp_valid_16) && _tmp_ready_8 && _tmp_valid_8) begin
        _tmp_data_16 <= _tmp_data_8;
      end 
      if(_tmp_valid_16 && _tmp_ready_16) begin
        _tmp_valid_16 <= 0;
      end 
      if((_tmp_ready_16 || !_tmp_valid_16) && _tmp_ready_8) begin
        _tmp_valid_16 <= _tmp_valid_8;
      end 
      if((_tmp_ready_17 || !_tmp_valid_17) && _tmp_ready_9 && _tmp_valid_9) begin
        _tmp_data_17 <= _tmp_data_9;
      end 
      if(_tmp_valid_17 && _tmp_ready_17) begin
        _tmp_valid_17 <= 0;
      end 
      if((_tmp_ready_17 || !_tmp_valid_17) && _tmp_ready_9) begin
        _tmp_valid_17 <= _tmp_valid_9;
      end 
      if((_tmp_ready_18 || !_tmp_valid_18) && _tmp_ready_10 && _tmp_valid_10) begin
        _tmp_data_18 <= _tmp_data_10;
      end 
      if(_tmp_valid_18 && _tmp_ready_18) begin
        _tmp_valid_18 <= 0;
      end 
      if((_tmp_ready_18 || !_tmp_valid_18) && _tmp_ready_10) begin
        _tmp_valid_18 <= _tmp_valid_10;
      end 
      if((_tmp_ready_19 || !_tmp_valid_19) && _tmp_ready_11 && _tmp_valid_11) begin
        _tmp_data_19 <= _tmp_data_11;
      end 
      if(_tmp_valid_19 && _tmp_ready_19) begin
        _tmp_valid_19 <= 0;
      end 
      if((_tmp_ready_19 || !_tmp_valid_19) && _tmp_ready_11) begin
        _tmp_valid_19 <= _tmp_valid_11;
      end 
      if((_tmp_ready_20 || !_tmp_valid_20) && _tmp_ready_12 && _tmp_valid_12) begin
        _tmp_data_20 <= _tmp_data_12;
      end 
      if(_tmp_valid_20 && _tmp_ready_20) begin
        _tmp_valid_20 <= 0;
      end 
      if((_tmp_ready_20 || !_tmp_valid_20) && _tmp_ready_12) begin
        _tmp_valid_20 <= _tmp_valid_12;
      end 
      if((_tmp_ready_21 || !_tmp_valid_21) && _tmp_ready_13 && _tmp_valid_13) begin
        _tmp_data_21 <= _tmp_data_13;
      end 
      if(_tmp_valid_21 && _tmp_ready_21) begin
        _tmp_valid_21 <= 0;
      end 
      if((_tmp_ready_21 || !_tmp_valid_21) && _tmp_ready_13) begin
        _tmp_valid_21 <= _tmp_valid_13;
      end 
      if((_tmp_ready_22 || !_tmp_valid_22) && _tmp_ready_14 && _tmp_valid_14) begin
        _tmp_data_22 <= _tmp_data_14;
      end 
      if(_tmp_valid_22 && _tmp_ready_22) begin
        _tmp_valid_22 <= 0;
      end 
      if((_tmp_ready_22 || !_tmp_valid_22) && _tmp_ready_14) begin
        _tmp_valid_22 <= _tmp_valid_14;
      end 
      if((_tmp_ready_23 || !_tmp_valid_23) && (_tmp_ready_15 && _tmp_ready_16) && (_tmp_valid_15 && _tmp_valid_16)) begin
        _tmp_data_23 <= $signed(_tmp_data_15) + $signed(_tmp_data_16);
      end 
      if(_tmp_valid_23 && _tmp_ready_23) begin
        _tmp_valid_23 <= 0;
      end 
      if((_tmp_ready_23 || !_tmp_valid_23) && (_tmp_ready_15 && _tmp_ready_16)) begin
        _tmp_valid_23 <= _tmp_valid_15 && _tmp_valid_16;
      end 
      if((_tmp_ready_24 || !_tmp_valid_24) && _tmp_ready_17 && _tmp_valid_17) begin
        _tmp_data_24 <= _tmp_data_17;
      end 
      if(_tmp_valid_24 && _tmp_ready_24) begin
        _tmp_valid_24 <= 0;
      end 
      if((_tmp_ready_24 || !_tmp_valid_24) && _tmp_ready_17) begin
        _tmp_valid_24 <= _tmp_valid_17;
      end 
      if((_tmp_ready_25 || !_tmp_valid_25) && _tmp_ready_18 && _tmp_valid_18) begin
        _tmp_data_25 <= _tmp_data_18;
      end 
      if(_tmp_valid_25 && _tmp_ready_25) begin
        _tmp_valid_25 <= 0;
      end 
      if((_tmp_ready_25 || !_tmp_valid_25) && _tmp_ready_18) begin
        _tmp_valid_25 <= _tmp_valid_18;
      end 
      if((_tmp_ready_26 || !_tmp_valid_26) && _tmp_ready_19 && _tmp_valid_19) begin
        _tmp_data_26 <= _tmp_data_19;
      end 
      if(_tmp_valid_26 && _tmp_ready_26) begin
        _tmp_valid_26 <= 0;
      end 
      if((_tmp_ready_26 || !_tmp_valid_26) && _tmp_ready_19) begin
        _tmp_valid_26 <= _tmp_valid_19;
      end 
      if((_tmp_ready_27 || !_tmp_valid_27) && _tmp_ready_20 && _tmp_valid_20) begin
        _tmp_data_27 <= _tmp_data_20;
      end 
      if(_tmp_valid_27 && _tmp_ready_27) begin
        _tmp_valid_27 <= 0;
      end 
      if((_tmp_ready_27 || !_tmp_valid_27) && _tmp_ready_20) begin
        _tmp_valid_27 <= _tmp_valid_20;
      end 
      if((_tmp_ready_28 || !_tmp_valid_28) && _tmp_ready_21 && _tmp_valid_21) begin
        _tmp_data_28 <= _tmp_data_21;
      end 
      if(_tmp_valid_28 && _tmp_ready_28) begin
        _tmp_valid_28 <= 0;
      end 
      if((_tmp_ready_28 || !_tmp_valid_28) && _tmp_ready_21) begin
        _tmp_valid_28 <= _tmp_valid_21;
      end 
      if((_tmp_ready_29 || !_tmp_valid_29) && _tmp_ready_22 && _tmp_valid_22) begin
        _tmp_data_29 <= _tmp_data_22;
      end 
      if(_tmp_valid_29 && _tmp_ready_29) begin
        _tmp_valid_29 <= 0;
      end 
      if((_tmp_ready_29 || !_tmp_valid_29) && _tmp_ready_22) begin
        _tmp_valid_29 <= _tmp_valid_22;
      end 
      if((_tmp_ready_30 || !_tmp_valid_30) && (_tmp_ready_23 && _tmp_ready_24) && (_tmp_valid_23 && _tmp_valid_24)) begin
        _tmp_data_30 <= $signed(_tmp_data_23) + $signed(_tmp_data_24);
      end 
      if(_tmp_valid_30 && _tmp_ready_30) begin
        _tmp_valid_30 <= 0;
      end 
      if((_tmp_ready_30 || !_tmp_valid_30) && (_tmp_ready_23 && _tmp_ready_24)) begin
        _tmp_valid_30 <= _tmp_valid_23 && _tmp_valid_24;
      end 
      if((_tmp_ready_31 || !_tmp_valid_31) && _tmp_ready_25 && _tmp_valid_25) begin
        _tmp_data_31 <= _tmp_data_25;
      end 
      if(_tmp_valid_31 && _tmp_ready_31) begin
        _tmp_valid_31 <= 0;
      end 
      if((_tmp_ready_31 || !_tmp_valid_31) && _tmp_ready_25) begin
        _tmp_valid_31 <= _tmp_valid_25;
      end 
      if((_tmp_ready_32 || !_tmp_valid_32) && _tmp_ready_26 && _tmp_valid_26) begin
        _tmp_data_32 <= _tmp_data_26;
      end 
      if(_tmp_valid_32 && _tmp_ready_32) begin
        _tmp_valid_32 <= 0;
      end 
      if((_tmp_ready_32 || !_tmp_valid_32) && _tmp_ready_26) begin
        _tmp_valid_32 <= _tmp_valid_26;
      end 
      if((_tmp_ready_33 || !_tmp_valid_33) && _tmp_ready_27 && _tmp_valid_27) begin
        _tmp_data_33 <= _tmp_data_27;
      end 
      if(_tmp_valid_33 && _tmp_ready_33) begin
        _tmp_valid_33 <= 0;
      end 
      if((_tmp_ready_33 || !_tmp_valid_33) && _tmp_ready_27) begin
        _tmp_valid_33 <= _tmp_valid_27;
      end 
      if((_tmp_ready_34 || !_tmp_valid_34) && _tmp_ready_28 && _tmp_valid_28) begin
        _tmp_data_34 <= _tmp_data_28;
      end 
      if(_tmp_valid_34 && _tmp_ready_34) begin
        _tmp_valid_34 <= 0;
      end 
      if((_tmp_ready_34 || !_tmp_valid_34) && _tmp_ready_28) begin
        _tmp_valid_34 <= _tmp_valid_28;
      end 
      if((_tmp_ready_35 || !_tmp_valid_35) && _tmp_ready_29 && _tmp_valid_29) begin
        _tmp_data_35 <= _tmp_data_29;
      end 
      if(_tmp_valid_35 && _tmp_ready_35) begin
        _tmp_valid_35 <= 0;
      end 
      if((_tmp_ready_35 || !_tmp_valid_35) && _tmp_ready_29) begin
        _tmp_valid_35 <= _tmp_valid_29;
      end 
      if((_tmp_ready_36 || !_tmp_valid_36) && (_tmp_ready_30 && _tmp_ready_31) && (_tmp_valid_30 && _tmp_valid_31)) begin
        _tmp_data_36 <= $signed(_tmp_data_30) + $signed(_tmp_data_31);
      end 
      if(_tmp_valid_36 && _tmp_ready_36) begin
        _tmp_valid_36 <= 0;
      end 
      if((_tmp_ready_36 || !_tmp_valid_36) && (_tmp_ready_30 && _tmp_ready_31)) begin
        _tmp_valid_36 <= _tmp_valid_30 && _tmp_valid_31;
      end 
      if((_tmp_ready_37 || !_tmp_valid_37) && _tmp_ready_32 && _tmp_valid_32) begin
        _tmp_data_37 <= _tmp_data_32;
      end 
      if(_tmp_valid_37 && _tmp_ready_37) begin
        _tmp_valid_37 <= 0;
      end 
      if((_tmp_ready_37 || !_tmp_valid_37) && _tmp_ready_32) begin
        _tmp_valid_37 <= _tmp_valid_32;
      end 
      if((_tmp_ready_38 || !_tmp_valid_38) && _tmp_ready_33 && _tmp_valid_33) begin
        _tmp_data_38 <= _tmp_data_33;
      end 
      if(_tmp_valid_38 && _tmp_ready_38) begin
        _tmp_valid_38 <= 0;
      end 
      if((_tmp_ready_38 || !_tmp_valid_38) && _tmp_ready_33) begin
        _tmp_valid_38 <= _tmp_valid_33;
      end 
      if((_tmp_ready_39 || !_tmp_valid_39) && _tmp_ready_34 && _tmp_valid_34) begin
        _tmp_data_39 <= _tmp_data_34;
      end 
      if(_tmp_valid_39 && _tmp_ready_39) begin
        _tmp_valid_39 <= 0;
      end 
      if((_tmp_ready_39 || !_tmp_valid_39) && _tmp_ready_34) begin
        _tmp_valid_39 <= _tmp_valid_34;
      end 
      if((_tmp_ready_40 || !_tmp_valid_40) && _tmp_ready_35 && _tmp_valid_35) begin
        _tmp_data_40 <= _tmp_data_35;
      end 
      if(_tmp_valid_40 && _tmp_ready_40) begin
        _tmp_valid_40 <= 0;
      end 
      if((_tmp_ready_40 || !_tmp_valid_40) && _tmp_ready_35) begin
        _tmp_valid_40 <= _tmp_valid_35;
      end 
      if((_tmp_ready_41 || !_tmp_valid_41) && (_tmp_ready_36 && _tmp_ready_37) && (_tmp_valid_36 && _tmp_valid_37)) begin
        _tmp_data_41 <= $signed(_tmp_data_36) + $signed(_tmp_data_37);
      end 
      if(_tmp_valid_41 && _tmp_ready_41) begin
        _tmp_valid_41 <= 0;
      end 
      if((_tmp_ready_41 || !_tmp_valid_41) && (_tmp_ready_36 && _tmp_ready_37)) begin
        _tmp_valid_41 <= _tmp_valid_36 && _tmp_valid_37;
      end 
      if((_tmp_ready_42 || !_tmp_valid_42) && _tmp_ready_38 && _tmp_valid_38) begin
        _tmp_data_42 <= _tmp_data_38;
      end 
      if(_tmp_valid_42 && _tmp_ready_42) begin
        _tmp_valid_42 <= 0;
      end 
      if((_tmp_ready_42 || !_tmp_valid_42) && _tmp_ready_38) begin
        _tmp_valid_42 <= _tmp_valid_38;
      end 
      if((_tmp_ready_43 || !_tmp_valid_43) && _tmp_ready_39 && _tmp_valid_39) begin
        _tmp_data_43 <= _tmp_data_39;
      end 
      if(_tmp_valid_43 && _tmp_ready_43) begin
        _tmp_valid_43 <= 0;
      end 
      if((_tmp_ready_43 || !_tmp_valid_43) && _tmp_ready_39) begin
        _tmp_valid_43 <= _tmp_valid_39;
      end 
      if((_tmp_ready_44 || !_tmp_valid_44) && _tmp_ready_40 && _tmp_valid_40) begin
        _tmp_data_44 <= _tmp_data_40;
      end 
      if(_tmp_valid_44 && _tmp_ready_44) begin
        _tmp_valid_44 <= 0;
      end 
      if((_tmp_ready_44 || !_tmp_valid_44) && _tmp_ready_40) begin
        _tmp_valid_44 <= _tmp_valid_40;
      end 
      if((_tmp_ready_45 || !_tmp_valid_45) && (_tmp_ready_41 && _tmp_ready_42) && (_tmp_valid_41 && _tmp_valid_42)) begin
        _tmp_data_45 <= $signed(_tmp_data_41) + $signed(_tmp_data_42);
      end 
      if(_tmp_valid_45 && _tmp_ready_45) begin
        _tmp_valid_45 <= 0;
      end 
      if((_tmp_ready_45 || !_tmp_valid_45) && (_tmp_ready_41 && _tmp_ready_42)) begin
        _tmp_valid_45 <= _tmp_valid_41 && _tmp_valid_42;
      end 
      if((_tmp_ready_46 || !_tmp_valid_46) && _tmp_ready_43 && _tmp_valid_43) begin
        _tmp_data_46 <= _tmp_data_43;
      end 
      if(_tmp_valid_46 && _tmp_ready_46) begin
        _tmp_valid_46 <= 0;
      end 
      if((_tmp_ready_46 || !_tmp_valid_46) && _tmp_ready_43) begin
        _tmp_valid_46 <= _tmp_valid_43;
      end 
      if((_tmp_ready_47 || !_tmp_valid_47) && _tmp_ready_44 && _tmp_valid_44) begin
        _tmp_data_47 <= _tmp_data_44;
      end 
      if(_tmp_valid_47 && _tmp_ready_47) begin
        _tmp_valid_47 <= 0;
      end 
      if((_tmp_ready_47 || !_tmp_valid_47) && _tmp_ready_44) begin
        _tmp_valid_47 <= _tmp_valid_44;
      end 
      if((_tmp_ready_48 || !_tmp_valid_48) && (_tmp_ready_45 && _tmp_ready_46) && (_tmp_valid_45 && _tmp_valid_46)) begin
        _tmp_data_48 <= $signed(_tmp_data_45) + $signed(_tmp_data_46);
      end 
      if(_tmp_valid_48 && _tmp_ready_48) begin
        _tmp_valid_48 <= 0;
      end 
      if((_tmp_ready_48 || !_tmp_valid_48) && (_tmp_ready_45 && _tmp_ready_46)) begin
        _tmp_valid_48 <= _tmp_valid_45 && _tmp_valid_46;
      end 
      if((_tmp_ready_49 || !_tmp_valid_49) && _tmp_ready_47 && _tmp_valid_47) begin
        _tmp_data_49 <= _tmp_data_47;
      end 
      if(_tmp_valid_49 && _tmp_ready_49) begin
        _tmp_valid_49 <= 0;
      end 
      if((_tmp_ready_49 || !_tmp_valid_49) && _tmp_ready_47) begin
        _tmp_valid_49 <= _tmp_valid_47;
      end 
      if((_tmp_ready_50 || !_tmp_valid_50) && (_tmp_ready_48 && _tmp_ready_49) && (_tmp_valid_48 && _tmp_valid_49)) begin
        _tmp_data_50 <= $signed(_tmp_data_48) + $signed(_tmp_data_49);
      end 
      if(_tmp_valid_50 && _tmp_ready_50) begin
        _tmp_valid_50 <= 0;
      end 
      if((_tmp_ready_50 || !_tmp_valid_50) && (_tmp_ready_48 && _tmp_ready_49)) begin
        _tmp_valid_50 <= _tmp_valid_48 && _tmp_valid_49;
      end 
    end
  end


endmodule



module multiplier_0
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_0
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_0
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_1
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_1
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_1
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_2
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_2
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_2
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_3
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_3
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_3
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_4
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_4
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_4
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_5
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_5
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_5
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_6
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_6
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_6
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_7
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_7
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_7
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule



module multiplier_8
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_8
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_8
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [14-1:0] b,
  output [46-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [14-1:0] _b;
  reg signed [46-1:0] _tmpval0;
  reg signed [46-1:0] _tmpval1;
  reg signed [46-1:0] _tmpval2;
  reg signed [46-1:0] _tmpval3;
  reg signed [46-1:0] _tmpval4;
  wire signed [46-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = dataflow_stencil.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *

def mkAdd(numports=8, mode=1):
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    
    inputs = [ (m.Input('d'+str(i), 32), m.Input('v'+str(i)), m.Output('r'+str(i)))
               for i in range(numports) ]
    
    dz = m.Output('dz', 32)
    vz = m.Output('vz')
    rz = m.Input('rz')
    
    df = Pipeline(m, 'df', clk, rst)

    dfvars = [ df.input(*p) for p in inputs ]

    if mode == 0:
        vars = dfvars
        for depth in range(int(math.log(numports,2))):
            tmp_vars = []
            for i in range(numports>>depth+1):
                tmp_vars.append( df(vars[i*2] + vars[i*2+1]) )
            vars = tmp_vars
        vars[0].output(dz, valid=vz, ready=rz)

    else:
        ret = None
        for v in dfvars:
            if ret is None:
                ret = v
                continue
            ret = df(ret + v)
        ret.output(dz, valid=vz, ready=rz)
    
    df.make_always()
    
    try:
        df.draw_graph()
    except:
        print('Dataflow graph could not be generated.', file=sys.stderr)

    return m

def mkTest(numports=8):
    m = Module('test')
    
    # target instance
    madd = mkAdd(numports)
    
    # copy paras and ports
    params = m.copy_params(madd)
    ports = m.copy_sim_ports(madd)

    clk = ports['CLK']
    rst = ports['RST']

    inputs = [ (ports['d'+str(i)], ports['v'+str(i)], ports['r'+str(i)])
               for i in range(numports) ]

    dz = ports['dz']
    vz = ports['vz']
    rz = ports['rz']
    
    uut = m.Instance(madd, 'uut',
                     params=m.connect_params(madd),
                     ports=m.connect_ports(madd))

    reset_done = m.Reg('reset_done', initval=0)
    reset_stmt = []
    reset_stmt.append( reset_done(0) )

    for p in inputs:
        reset_stmt.extend( [ p[0](0), p[1](0) ] )

    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000),
        reset_done(1),
        nclk(clk),
        Delay(10000),
        Systask('finish'),
    )

    for i,p in enumerate(inputs):
        count = m.TmpReg(32, initval=0)
        d = p[0]
        v = p[1]
        r = p[2]
        fsm = FSM(m, 'fsm'+str(i), clk, rst)
        fsm.add(v(0))
        fsm.goto_next(cond=reset_done)
        fsm.add(v(1))
        fsm.goto_next()
        fsm.add(d(d + i + 1), cond=r)
        fsm.add(count.inc(), cond=r)
        fsm.goto_next(cond=AndList(count==10, r))
        fsm.add(v(0))
        fsm.make_always()

    zfsm = FSM(m, 'zfsm', clk, rst)
    zfsm.add(rz(0))
    zfsm.goto_next(cond=reset_done)
    zfsm.goto_next()
    zinit= zfsm.current
    zfsm.add(rz(1), cond=vz)
    zfsm.goto_next(cond=vz)
    for i in range(10):
        zfsm.add(rz(0))
        zfsm.goto_next()
    zfsm.goto(zinit)
    zfsm.make_always()

    disp = []
    for p in inputs:
        d = p[0]
        v = p[1]
        r = p[2]

        disp.append( If(AndList(v, r))( Systask('display', d.name + '=%d', d) ) )

    m.Always(Posedge(clk))(
        If(reset_done)(
            disp,
            If(AndList(vz, rz))(
                Systask('display', dz.name + '=%d', dz)
            )))
    
    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    #sim.view_waveform()
